/*
 
 Copyright (c) 2011-2013 Ethan Funk
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
 documentation files (the "Software"), to deal in the Software without restriction, including without limitation 
 the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, 
 and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in all copies or substantial portions 
 of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
 TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF 
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
 IN THE SOFTWARE.
 
*/

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <fcntl.h>
#include <ctype.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <grp.h>
#include <string.h>
#include <signal.h>
#include "rs.h"
#include "rsp.h"

/*
 #include <math.h>
 #include <sys/time.h>
*/

#ifdef __APPLE__
	#include <malloc/malloc.h>
#endif

#define clientID "simpleEncoder 0.3"

struct destination {
	struct destination	*next;
	struct sockaddr_in6 address;
	int socket;						// will be sent either to a common (reused) udp socket or a custom 
									// socket with proper ttl set if the destination is a multicast group
};

struct encoderContext {
	cJSON *conf;
	int meta_socket;
	int rsp_socket4;
	int rsp_socket6;
	struct rspSession *rsp;
	struct destination *destinations;
};

cJSON *readConfigFile(FILE * fd)
{
	long len;
	char *data;
	cJSON *root_conf;
	
	root_conf = NULL;
	
	fseek(fd, 0, SEEK_END);
	len = ftell(fd);
	fseek(fd, 0, SEEK_SET);
	
	data = (char *)malloc(len + 1);
	data[len] = 0;
	fread(data, 1, len, fd);
	root_conf = cJSON_Parse(data);
	free(data);	
	return root_conf;
}

void rspSessionSetup(struct encoderContext *cPtr)
{
	unsigned char rspErr;
	cJSON *item, *group;
	struct rspSession *session;
	
	session = rspSessionNew(clientID);
	// detach rsp object from encoder config and re-added by refernce so the the rsp session will own the object
	if((group = cJSON_GetObjectItem(cPtr->conf, "rsp")) == NULL)
		goto fail;
	item = group;
	if(rspSessionConfigNextJSON(session, &item) == RSP_ERROR_NONE){
		if(group = cJSON_GetObjectItem(group, "Format")){
			if(item = cJSON_GetObjectItem(group, "PrivateKeyFile")){
				if(item->valuestring && strlen(item->valuestring)){
					FILE *keyFile = fopen(item->valuestring, "r");				// Open the specified private key file
					if(keyFile == NULL)
						goto fail;
					rspErr = rspSessionSetPrivKeyFile(session, keyFile, "");	
					fclose(keyFile);										
					if(rspErr)
						goto fail;
				}
			}
		}
	}
	
	rspSessionClear(session, TRUE);
	if(rspSessionInit(session) == RSP_ERROR_NONE){
		cPtr->rsp = session;
		return;
	}
	   
fail:
	rspSessionFree(session);
	cPtr->rsp = NULL;
}

struct sockaddr_in6 *setSockAddr(struct sockaddr_in6 *adrPtr, unsigned char ip6, unsigned short port, const char *addr)
{
	unsigned int size;
	struct sockaddr_in *v4bindAddr;
	struct in6_addr v6bindto;
	u_int32_t v4bindto;
	
	if(ip6){
		// IPv6 network settings
		size = sizeof(struct sockaddr_in6);
		bzero(adrPtr, size);
#ifndef __linux__
		adrPtr->sin6_len = sizeof(struct sockaddr_in6);
#endif		
		if(inet_pton(AF_INET6, addr, &v6bindto) <= 0)
			return NULL;
		adrPtr->sin6_family = AF_INET6;
		adrPtr->sin6_addr = v6bindto;
		adrPtr->sin6_port = htons(port);
	}else{
		size = sizeof(struct sockaddr_in6);
		bzero(adrPtr, size);
		v4bindAddr = (struct sockaddr_in *)adrPtr;
#ifndef __linux__
		v4bindAddr->sin_len = sizeof(struct sockaddr_in);
#endif		
		if(inet_pton(AF_INET, addr, &v4bindto) <= 0)
			return NULL;
		v4bindAddr->sin_family = AF_INET;
		v4bindAddr->sin_addr.s_addr = v4bindto;
		v4bindAddr->sin_port = htons(port);
	}
	return adrPtr;
}

int socketSetup(struct encoderContext *cPtr, unsigned char ttl, unsigned char ip6)
{
	cJSON *group, *item;
	int sd;
	unsigned int size;
	char *bindTo;
	struct sockaddr_in6 bindAddr;
	int trueVal = 1;
	
	sd = -1;
	bindTo = NULL;
	if(ip6){
		if((sd = socket(AF_INET6, SOCK_DGRAM, IPPROTO_IP)) < 0)
			return -1;
		setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &trueVal, sizeof(trueVal));
		if(group = cJSON_GetObjectItem(cPtr->conf, "IP6")){
			// IPv6 network settings found
			if(item = cJSON_GetObjectItem(group, "Bind"))
				bindTo = item->valuestring;
			if(bindTo == NULL)
				goto fail;
			if(item = cJSON_GetObjectItem(group, "diffServHex")){
				if(item->valuestring && strlen(item->valuestring)){
					int dsval;
					dsval = strtoul(item->valuestring, NULL, 16);
					dsval = dsval & 0xFB;  // upper six bits only are valid
					if(setsockopt(sd, IPPROTO_IP, IP_TOS, &dsval, sizeof(dsval)) < 0)
						goto fail;
				}
			}
			if(setSockAddr(&bindAddr, TRUE, 0, bindTo) == NULL)
				goto fail;
			size = sizeof(struct sockaddr_in6);
			if(bind(sd, (struct sockaddr *)&bindAddr, size) < 0)
				goto fail;
		}
	}else{
		// use old (IPv4) network settings
		if((sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0)
			return -1;
		setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &trueVal, sizeof(trueVal));
		if(group = cJSON_GetObjectItem(cPtr->conf, "IP4")){
			if(item = cJSON_GetObjectItem(group, "Bind"))
				bindTo = item->valuestring;
			if(bindTo == NULL)
				goto fail;
			if(setSockAddr(&bindAddr, FALSE, 0, bindTo) == NULL)
				goto fail;
			if(item = cJSON_GetObjectItem(group, "diffServHex")){
				if(item->valuestring && strlen(item->valuestring)){
					int dsval;
					dsval = strtoul(item->valuestring, NULL, 16);
					dsval = dsval & 0xFB;  // upper six bits only are valid
					if(setsockopt(sd, IPPROTO_IP, IP_TOS, &dsval, sizeof(dsval)) < 0)
						goto fail;
				}
			}
			size = sizeof(struct sockaddr_in);
			if(bind(sd, (struct sockaddr *)&bindAddr, size) < 0)
				goto fail;
		}
	}
	if(ttl){
		// set up multicast properties
		unsigned char value = 0;
		if(setsockopt(sd, IPPROTO_IP, IP_MULTICAST_LOOP, (unsigned char *)&value, sizeof(value)) < 0){
			goto fail;
		}
		if(bindTo){
			// note: bind to the same interface the relay socket is bound to...
			if(setsockopt(sd, IPPROTO_IP, IP_MULTICAST_IF, (char *)&bindAddr, sizeof(bindAddr)) < 0){
				goto fail;
			}
		}
		// multicast variable is zero for no multicast, otherwise it contains the TTL value for the multicast hop threshold
		if(setsockopt(sd, IPPROTO_IP, IP_MULTICAST_TTL, (unsigned char *)&ttl, sizeof(ttl)) < 0){
			goto fail;
		}
	}
	return sd;
	
fail:
	if(sd >= 0)
		close(sd);
	return -1;
}

void addSendtoDestination(cJSON *listener, struct encoderContext *cPtr)
{
	// add static unicast or multicast item
	cJSON *item, *type;
	unsigned char ttl, ip6;
	unsigned short portNo;
	struct destination *record;
	
	portNo = 0;
	ttl = 0;
	ip6 = FALSE;
	
	if(item = cJSON_GetObjectItem(listener, "MulticastTTL"))
		ttl = item->valueint;
	if(item = cJSON_GetObjectItem(listener, "Port"))
		portNo = item->valueint;						
	if(portNo && (item = cJSON_GetObjectItem(listener, "Address"))){
		if((item->valuestring == NULL) || (strlen(item->valuestring) == 0))
			portNo = 0;
	}
	if(portNo){
		struct sockaddr_in6 sockAddr;
		struct sockaddr_in6 *addrPtr;
		
		if((type = cJSON_GetObjectItem(listener, "Type")) && type->valuestring && (strcasecmp(type->valuestring, "IP6") == 0))
			ip6 = TRUE;
		if(addrPtr = setSockAddr(&sockAddr, ip6, portNo, item->valuestring)){
			if((record = (struct destination *)malloc(sizeof(struct destination))) == NULL){
				fprintf(stderr, "Failed to allocate memory for sendto record.\n");
				return;
			}
			record->socket = -1;
			if(ttl){
				// create a multicast socket
				if(record->socket = socketSetup(cPtr, ttl, ip6) < 0){
					// socket failure
					free(record);
					fprintf(stderr, "Failed to set multicast ttl on static listener %s on port %u: Socket failure.\n", item->valuestring, portNo);
					return;
				}
			}else{
				// use or set up an existing unicast socket
				if(ip6){
					if(cPtr->rsp_socket6 < 0){
						// create new....
						if((cPtr->rsp_socket6 = socketSetup(cPtr, 0, ip6)) < 0){
							free(record);
							fprintf(stderr, "Failed to create outgoing ip6 unicast socket.\n");
							return;
						}
					}
					// use existing...
					record->socket = cPtr->rsp_socket6;
				}else{
					if(cPtr->rsp_socket4 < 0){
						// create new....
						if((cPtr->rsp_socket4 = socketSetup(cPtr, 0, ip6)) < 0){
							free(record);
							fprintf(stderr, "Failed to create outgoing ip4 unicast socket.\n");
							return;
						}
					}
					// use existing...
					record->socket = cPtr->rsp_socket4;
				}
			}
			record->address = *addrPtr;
			record->next = cPtr->destinations;
			cPtr->destinations = record;			
			return;
		}
	}
	if(item && item->valuestring)
		fprintf(stderr, "Failed to add static listener %s on port %u: bad address or port number.\n", item->valuestring, portNo);
	else 
		fprintf(stderr, "Failed to add unknown static listener on port %u: bad address or port number.\n", portNo);
}

unsigned char queueFormatMetadata(struct encoderContext *cPtr)
{
	cJSON *group, *item, *root;
	char *str;

	// make sure there is a Content section in the config jSON object
	if((group = cJSON_GetObjectItem(cPtr->conf, "rsp")) == NULL)
		return FALSE;
	if((group = cJSON_GetObjectItem(group, "Content")) == NULL)
		return FALSE;
	
	// make sure the content section has at least a type parameter
	if((item = cJSON_GetObjectItem(group, "Type")) == NULL)
		return FALSE;
	if(!item->valuestring || (strlen(item->valuestring) == 0))
		return FALSE;
	
	// make a copy of the object
	if((str = cJSON_PrintUnformatted(group)) == NULL)
	   return FALSE;
	if((group = cJSON_Parse(str)) == NULL){
		free(str);
		return FALSE;
	}

	// add mID field to make this a repeate item
	srand(time(NULL));
	unsigned int mIDVal;
	while(((mIDVal = rand()) & 0xffff) == 0);
	cJSON_AddNumberToObject(group, "mID", mIDVal);
	
	// and queue it for RSP transmission
	root =  cJSON_CreateObject();
	cJSON_AddItemToObject(root, "Content", group);
	rspSessionQueueMetadata(cPtr->rsp, root, NULL);
	return TRUE;
}

void queueMeta(const char *data, struct encoderContext *cPtr)
{
	cJSON *obj;
	
	if(obj = cJSON_Parse(data))
		rspSessionQueueMetadata(cPtr->rsp, obj, NULL);
	else{
		fprintf(stderr, "received malformed jSON metadata\n");
		syslog(LOG_NOTICE, "received malformed jSON metadata");
	}
}

int	setupMetadataFIFO(struct encoderContext *cPtr)
{
	cJSON *item, *group;

	if((item = cJSON_GetObjectItem(cPtr->conf, "metaFIFO")) == NULL)
		return -1;
	if(!item->valuestring || (strlen(item->valuestring) == 0))
		return -1;
	if(mkfifo(item->valuestring, 0622)){
		if(errno != EEXIST)
			// error is not "already exists"
			return -1;
	}
	   
	// if a named socket group is specified, set the socket group-id to the named group
	if(group = cJSON_GetObjectItem(cPtr->conf, "FIFOGroup")){
	   if(group && group->valuestring && strlen(group->valuestring)){
		   struct group grp, *tempGrpPtr;
		   struct group *grpptr;
		   int  grplinelen, rcode;
		   char *grpbuffer;
		   
		   grpptr = &grp;
		   grplinelen = 1024;
		   grpbuffer = (char *)malloc(grplinelen);
		   while((rcode = getgrnam_r(group->valuestring, grpptr, grpbuffer, grplinelen, &tempGrpPtr)) == ERANGE){
			   // need a bigger buffer...
			   grplinelen = grplinelen + 1024;
			   if((grpbuffer = realloc(grpbuffer, grplinelen)) == NULL)
				   break;
		   }
		   if((rcode == 0) && tempGrpPtr)
			   rcode = chown(item->valuestring, getuid(), grp.gr_gid);
		   if(rcode)
			   fprintf(stderr, "Failed to set meta data FIFO permission group to %s. Continuing...\n", item->valuestring);
		   else 
			   // set socket permission to write owner and read group. Everyone else, no access
			   chmod(item->valuestring, 0620);
		   free(grpbuffer);				
	   }
	}
	return open(item->valuestring, O_RDWR); 
}

unsigned char configure(FILE * fd, struct encoderContext *cPtr)
{
	cJSON *item, *group, *sendto;
	int socket4, socket6;
	
	socket4 = -1;
	socket6 = -1;
	
	if((cPtr->conf = readConfigFile(fd)) == NULL){
		fprintf(stderr, "Configuration failed: Configuration file contains a jSON format error.\n");
		goto fail;
	}
	if((cPtr->meta_socket = setupMetadataFIFO(cPtr)) < 0){
		fprintf(stderr, "Configuration failed: Failed to configure meta data FIFO.\n");
		goto fail;
	}
	
	rspSessionSetup(cPtr);
	if(cPtr->rsp == NULL){
		fprintf(stderr, "Configuration failed: couldn't to configure the rsp format settings.\n");
		goto fail;
	}
	
	if(!queueFormatMetadata(cPtr)){
		fprintf(stderr, "Configuration failed: ininvalid Content settings.\n");
		goto fail;
	}
			
		
	if(((group = cJSON_GetObjectItem(cPtr->conf, "rsp")) == NULL) || ((sendto = cJSON_GetObjectItem(group, "sendto")) == NULL)){
		fprintf(stderr, "Configuration failed: missing send-to addresses.\n");
		goto fail;
	}

	item = sendto->child;
	while(item){
		addSendtoDestination(item, cPtr);
		item = item->next;
	}
	if(cPtr->destinations == NULL){
		fprintf(stderr, "Configuration failed: no valid send-to addresses.\n");
		goto fail;
	}
	
	return TRUE;

fail:
	return FALSE;
}

void io_loop(struct encoderContext *cPtr)
{
	struct pollfd fds[2];
	char read_buf[256];
	char *metaBuf;
	char *frag, *tmp_ptr, *curMetaStr;
	unsigned char *packetPtr;
	int n;
	unsigned int metaPos, dataFrameSize, size, frag_offset;
	unsigned char dataFrame[255];
	struct destination *dest_ptr;
		
	metaBuf = NULL;
	curMetaStr = NULL;
	frag_offset = 0;
	
	fds[0].fd = STDIN_FILENO;
	fds[0].events = POLLIN;
	
	fds[1].fd = cPtr->meta_socket;
	fds[1].events = POLLIN;
	
	while(1){
		if(poll(fds, 2, -1) > 0){
			// check for metadata FIFO data availability
			if(fds[1].revents & (POLLIN | POLLERR)){
				if(fds[1].revents & POLLERR){
					// error on this file descriptor
					return;
				}else{
					// new meta data to read
					if((n = read(fds[1].fd, read_buf, sizeof(read_buf) - 1)) > 0){
						// null terminate the buffer
						read_buf[n] = 0;
						tmp_ptr = read_buf;
						// look for new line (\n)
						while(frag = strchr(tmp_ptr, '\n')){
							*frag = 0;
							// add the contents prior to the \n to the partial line buffer
							appendstr(&metaBuf, tmp_ptr);
							
							// process complete line
							queueMeta(metaBuf, cPtr);
							
							// clear the processed line buffer
							free(metaBuf);
							metaBuf = NULL;
							// move to next char in after the new line and continue
							tmp_ptr = frag + 1;
						}
						// add any left over to the partial line buffer for next time
						appendstr(&metaBuf, tmp_ptr);
					}else{
						// error on this file descriptor
						return;
					}
				}
			}
				
			// check for stream input data availability (STDIN)
			if(fds[0].revents & (POLLIN | POLLERR)){
				if(fds[0].revents & POLLERR){
					// error on this file descriptor
					return;
				}else{
					// new input data to encode
					if((n = read(fds[0].fd, read_buf, sizeof(read_buf))) > 0){
						tmp_ptr = read_buf;
						while(n > 0){
							// NOTE: First byte of frame is meta data, followed by audio data
							dataFrameSize = 255 - cPtr->rsp->FECroots;
							if(n < (signed int)((dataFrameSize - 1) - frag_offset)){	// note 1 extra byte reserved for meta data stream
								// copy fragment
								memcpy(dataFrame + frag_offset + 1, tmp_ptr, n);
								frag_offset = frag_offset + n;
								n = 0;
							}else{
								// enough to fill a frame
								size = (dataFrameSize - 1) - frag_offset;					// note 1 extra byte reserved for meta data stream
								memcpy(dataFrame + frag_offset + 1, tmp_ptr, size);
								frag_offset = 0;
								n = n - size;
								tmp_ptr = tmp_ptr + size;
								
								if(curMetaStr == NULL){
									// check to see if there is new metadata to start adding to frames
									if(curMetaStr = rspSessionNextMetaStr(cPtr->rsp))
										metaPos = 0;						 
								}
								
								// set next metadataByte, if any
								if(curMetaStr){
									if((dataFrame[0] = curMetaStr[metaPos++]) == 0){
										// end of string
										free(curMetaStr);
										curMetaStr = NULL;
									}
								}else
									// No metadata: set metadata byte in packet to NULL char
									dataFrame[0] = 0; 
								
								// we have a full frame... process through data RS encoder and write to interleaver row
								if(!rspSessionWriteData(cPtr->rsp, dataFrame, dataFrameSize)){
									
									// no error... check for packets comming out of interleaver to send
									while(size = rspSessionReadPacket(cPtr->rsp, &packetPtr, NULL)){
										// travers the destinations linked list, sending the packet to each destination.
										if(dest_ptr = cPtr->destinations){
											do{
												rspSendto(dest_ptr->socket, packetPtr, size, 0, (struct sockaddr*)&dest_ptr->address);
											}
											while(dest_ptr = dest_ptr->next);
										}else
											break;
									}
								}
							}
						}
						
					}else{
						// error on this file descriptor
						return;
					}
				}
			}
		}else{
			// poll error
			return;
		}
	}
}
	
void finishCleanup(struct encoderContext *cPtr)
{
	struct destination *dPtr, *prev;
	cJSON *item;
	
	if(cPtr->meta_socket >= 0){
		shutdown(cPtr->meta_socket, SHUT_RDWR);
		close(cPtr->meta_socket);
	}
	// remove FIFO
	if(cPtr->conf && (item = cJSON_GetObjectItem(cPtr->conf, "metaFIFO"))){
		if(item->valuestring && strlen(item->valuestring))
			remove(item->valuestring);
	}

	if(cPtr->rsp_socket6 >= 0){
		shutdown(cPtr->rsp_socket6, SHUT_RDWR);
		close(cPtr->rsp_socket6);
	}
	if(cPtr->rsp_socket4 >= 0){
		shutdown(cPtr->rsp_socket4, SHUT_RDWR);
		close(cPtr->rsp_socket4);
	}
	
	// close sendto sockets (incase they are multicast and different than the above
	// and free the corrisponding destination record
	dPtr = cPtr->destinations;
	while(prev = dPtr){
		shutdown(prev->socket, SHUT_RDWR);
		close(prev->socket);
		dPtr = prev->next;
		free(prev);
	}
	
	if(cPtr->conf)
		cJSON_Delete(cPtr->conf);
	if(cPtr->rsp)
		rspSessionFree(cPtr->rsp);	
}

int main(int argc, const char * argv[])
{
	FILE *confFile;
	struct encoderContext context; 

	if(argc == 2){
		// configure and run a encoder instance
	   
		// Block SIGPIPE
		sigblock(sigmask(SIGPIPE));
		
		// Open the specified configuration file
		if((confFile = fopen(argv[1], "r"))	== NULL){	
			syslog(LOG_ALERT, "Failed to open the specified configuration file; shutting down.");
			fprintf(stderr, "Failed to open the specified configuration file; shutting down.\n");
			return 0;
		}

		bzero(&context, sizeof(context));
		context.meta_socket = -1;
		context.rsp_socket4 = -1;
		context.rsp_socket6 = -1;
		if(configure(confFile, &context)){
			fclose(confFile);
			// run the input processing loop
			io_loop(&context);
		}else{
			fclose(confFile);
			syslog(LOG_ALERT, "Encoder configuration failed; shutting down.");
			fprintf(stderr, "Encoder configuration failed; shutting down.\n");
		}
		
		// the processing loop returned or config failed: cleanup and quit
		finishCleanup(&context);
		return 0;		
	
	}else{
		fprintf(stdout, "Usage:\n\n(some stream data source) | rspEncoder /encoder/config/file/path\n");
		fprintf(stdout, "\nStarts an instance of rspEncoder, configured using the specified configurstion file\n");
		fprintf(stdout, "in jSON format.  See example file (should have been included with this program).\n");
		fprintf(stdout, "\n\nThe encoder listens to stdin for streaming data to encode and listens to the named\n");
		fprintf(stdout, "\nFIFO (setup in the configuration file) for new jSON formated metadata to post to the stream.\n");
		fprintf(stdout, "\nMetadata must be sent to the named FIFO as a single line per metadata object, following\n");
		fprintf(stdout, "\nthe rsp metadata formating convention (see rsp specification file).  The single line per\n");
		fprintf(stdout, "\nmetadata object requirement implies non-printable (white space stripped) jSON format.\n\n");
		return 0;		
	}
}

